import pygame
from pygame.locals import *
from ..model.constants import *
from ..model.mouse import Mouse
from ..model.nodes import NodeGroup
from ..model.cheeses import CheeseGroup
from ..model.cats import CatGroup
from .pauser import Pause
from ..view.text import TextGroup
from ..view.sprites import LifeSprites
from ..view.sprites import MazeSprites
from ..model.mazedata import MazeData
from .events_manager import EventsManager
from .levels_manager import LevelManager
from ..view.game_view import GameView 

class GameController(object):
    def __init__(self):
        pygame.init()
        self.screen = pygame.display.set_mode(SCREENSIZE, 0, 32)
        self.background = None
        self.background_norm = None
        self.background_flash = None
        self.clock = pygame.time.Clock()
        self.pause = Pause(True)
        self.level = 0
        self.lives = 5
        self.score = 0
        self.textgroup = TextGroup()
        self.lifesprites = LifeSprites(self.lives)
        self.flashBG = False
        self.flashTime = 0.2
        self.flashTimer = 0
        self.mazedata = MazeData()
        self.events_manager = EventsManager(self)
        self.level_manager = LevelManager(self)
        self.view = GameView(self)

    def startGame(self):      
        self.mazedata.loadMaze(self.level)
        self.mazesprites = MazeSprites("CheeseChase/resources/"+self.mazedata.obj.name+".txt", "CheeseChase/resources/"+self.mazedata.obj.name+"_rotation.txt")
        self.setBackground()
        self.nodes = NodeGroup("CheeseChase/resources/"+self.mazedata.obj.name+".txt")
        self.mazedata.obj.setPortalPairs(self.nodes)
        self.mazedata.obj.connectHomeNodes(self.nodes)
        self.mouse = Mouse(self.nodes.getNodeFromTiles(*self.mazedata.obj.mouseStart))
        self.cheeses = CheeseGroup("CheeseChase/resources/"+self.mazedata.obj.name+".txt", self.mazesprites)
        self.cats = CatGroup(self.nodes.getStartTempNode(), self.mouse)

        self.cats.cat2.setStartNode(self.nodes.getNodeFromTiles(*self.mazedata.obj.addOffset(2, 3)))
        self.cats.cat3.setStartNode(self.nodes.getNodeFromTiles(*self.mazedata.obj.addOffset(0, 3)))
        self.cats.cat4.setStartNode(self.nodes.getNodeFromTiles(*self.mazedata.obj.addOffset(4, 3)))
        self.cats.setSpawnNode(self.nodes.getNodeFromTiles(*self.mazedata.obj.addOffset(2, 3)))
        self.cats.cat1.setStartNode(self.nodes.getNodeFromTiles(*self.mazedata.obj.addOffset(2, 0)))

        self.nodes.denyHomeAccess(self.mouse)
        self.nodes.denyHomeAccessList(self.cats)
        self.cats.cat3.startNode.denyAccess(RIGHT, self.cats.cat3)
        self.cats.cat4.startNode.denyAccess(LEFT, self.cats.cat4)
        self.mazedata.obj.denyCatsAccess(self.cats, self.nodes)

    def setBackground(self):
        self.background_norm = pygame.surface.Surface(SCREENSIZE).convert()
        self.background_norm.fill(WHITE)
        self.background_flash = pygame.surface.Surface(SCREENSIZE).convert()
        self.background_flash.fill(WHITE)
        self.background_norm = self.mazesprites.constructBackground(self.background_norm, self.level%5)
        self.background_flash = self.mazesprites.constructBackground(self.background_flash, 5)
        self.flashBG = False
        self.background = self.background_norm

    def update(self):
        dt = self.clock.tick(30) / 1000.0
        self.textgroup.update(dt)
        self.cheeses.update(dt)
        if not self.pause.paused:
            self.cats.update(dt)      
            self.events_manager.checkCheeseEvents()
            self.events_manager.checkCatEvents()

        if self.mouse.alive:
            if not self.pause.paused:
                self.mouse.update(dt)
        else:
            self.mouse.update(dt)

        if self.flashBG:
            self.flashTimer += dt
            if self.flashTimer >= self.flashTime:
                self.flashTimer = 0
                if self.background == self.background_norm:
                    self.background = self.background_flash
                else:
                    self.background = self.background_norm

        afterPauseMethod = self.pause.update(dt)
        if afterPauseMethod is not None:
            afterPauseMethod()
        self.events_manager.checkEvents() 
        self.view.render()

    def showEntities(self):
        self.mouse.visible = True
        self.cats.show()

    def hideEntities(self):
        self.mouse.visible = False
        self.cats.hide()

    def updateScore(self, points):
       self.score += points
       self.textgroup.updateScore(self.score)




from .constants import *

class MazeBase(object):
    def __init__(self):
        self.portalPairs = {}
        self.homeoffset = (0, 0)
        self.catNodeDeny = {UP:(), DOWN:(), LEFT:(), RIGHT:()}

    def setPortalPairs(self, nodes):
        for pair in list(self.portalPairs.values()):
            nodes.setPortalPair(*pair)

    def connectHomeNodes(self, nodes):
        key = nodes.createHomeNodes(*self.homeoffset)
        nodes.connectHomeNodes(key, self.homenodeconnectLeft, LEFT)
        nodes.connectHomeNodes(key, self.homenodeconnectRight, RIGHT)

    def addOffset(self, x, y):
        return x+self.homeoffset[0], y+self.homeoffset[1]

    def denyCatsAccess(self, cats, nodes):
        nodes.denyAccessList(*(self.addOffset(2, 3) + (LEFT, cats)))
        nodes.denyAccessList(*(self.addOffset(2, 3) + (RIGHT, cats)))

        for direction in list(self.catNodeDeny.keys()):
            for values in self.catNodeDeny[direction]:
                nodes.denyAccessList(*(values + (direction, cats)))


class Maze1(MazeBase):
    def __init__(self):
        MazeBase.__init__(self)
        self.name = "maze1"
        self.portalPairs = {0:((0, 17), (27, 17))}
        self.homeoffset = (11.5, 14)
        self.homenodeconnectLeft = (12, 14)
        self.homenodeconnectRight = (15, 14)
        self.mouseStart = (15, 26)
        self.catNodeDeny = {UP:((12, 14), (15, 14), (12, 26), (15, 26)), LEFT:(self.addOffset(2, 3),),
                              RIGHT:(self.addOffset(2, 3),)}


class Maze2(MazeBase):
    def __init__(self):
        MazeBase.__init__(self)
        self.name = "maze2"
        self.portalPairs = {0:((0, 4), (27, 4)), 1:((0, 26), (27, 26))}
        self.homeoffset = (11.5, 14)
        self.homenodeconnectLeft = (9, 14)
        self.homenodeconnectRight = (18, 14)
        self.mouseStart = (16, 26)
        self.catNodeDeny = {UP:((9, 14), (18, 14), (11, 23), (16, 23)), LEFT:(self.addOffset(2, 3),),
                              RIGHT:(self.addOffset(2, 3),)}


class MazeData(object):
    def __init__(self):
        self.obj = None
        self.mazedict = {0:Maze1, 1:Maze2}

    def loadMaze(self, level):
        self.obj = self.mazedict[level%len(self.mazedict)]()



from ..model.constants import *

class Animator(object):
    def __init__(self, frames=[], speed=20, loop=True):
        self.frames = frames
        self.current_frame = 0
        self.speed = speed
        self.loop = loop
        self.dt = 0
        self.finished = False

    def reset(self):
        self.current_frame = 0
        self.finished = False

    def update(self, dt):
        if not self.finished:
            self.nextFrame(dt)
        if self.current_frame == len(self.frames):
            if self.loop:
                self.current_frame = 0
            else:
                self.finished = True
                self.current_frame -= 1
   
        return self.frames[self.current_frame]

    def nextFrame(self, dt):
        self.dt += dt
        if self.dt >= (1.0 / self.speed):
            self.current_frame += 1
            self.dt = 0





                        



from .constants import *

class MainMode(object):
    def __init__(self):
        self.timer = 0
        self.scatter()

    def update(self, dt):
        self.timer += dt
        if self.timer >= self.time:
            if self.mode is SCATTER:
                self.chase()
            elif self.mode is CHASE:
                self.scatter()

    def scatter(self):
        self.mode = SCATTER
        self.time = 7
        self.timer = 0

    def chase(self):
        self.mode = CHASE
        self.time = 20
        self.timer = 0


class ModeController(object):
    def __init__(self, entity):
        self.timer = 0
        self.time = None
        self.mainmode = MainMode()
        self.current = self.mainmode.mode
        self.entity = entity 

    def update(self, dt):
        self.mainmode.update(dt)
        if self.current is FREIGHT:
            self.timer += dt
            if self.timer >= self.time:
                self.time = None
                self.entity.normalMode()
                self.current = self.mainmode.mode
        elif self.current in [SCATTER, CHASE]:
            self.current = self.mainmode.mode

        if self.current is SPAWN:
            if self.entity.node == self.entity.spawnNode:
                self.entity.normalMode()
                self.current = self.mainmode.mode

    def setFreightMode(self):
        if self.current in [SCATTER, CHASE]:
            self.timer = 0
            self.time = 7
            self.current = FREIGHT
        elif self.current is FREIGHT:
            self.timer = 0

    def setSpawnMode(self):
        if self.current is FREIGHT:
            self.current = SPAWN


import pygame
from pygame.locals import *
from ..model.constants import *

class GameView:
    def __init__(self, controller):
        self.controller = controller 

    def render(self):
        screen = self.controller.screen
        screen.blit(self.controller.background, (0, 0))
        self.controller.cheeses.render(screen)
        self.controller.mouse.render(screen)
        self.controller.cats.render(screen)
        self.controller.textgroup.render(screen)

        for i, img in enumerate(self.controller.lifesprites.images):
            x = img.get_width() * i
            y = SCREENHEIGHT - img.get_height()
            screen.blit(img, (x, y))
            
        pygame.display.update()

    


import pygame
from pygame.locals import *
from .vector import Vector2
from .constants import *
from random import randint

class Entity(object):
    def __init__(self, node):
        self.name = None
        self.directions = {UP:Vector2(0, -1),DOWN:Vector2(0, 1), 
                          LEFT:Vector2(-1, 0), RIGHT:Vector2(1, 0), STOP:Vector2()}
        self.direction = STOP
        self.setSpeed(100)
        self.radius = 10
        self.collideRadius = 5
        self.color = WHITE
        self.visible = True
        self.disablePortal = False
        self.goal = None
        self.directionMethod = self.randomDirection
        self.setStartNode(node)
        self.image = None

    def setPosition(self):
        self.position = self.node.position.copy()

    def update(self, dt):
        self.position += self.directions[self.direction]*self.speed*dt
         
        if self.overshotTarget():
            self.node = self.target
            directions = self.validDirections()
            direction = self.directionMethod(directions)
            if not self.disablePortal:
                if self.node.neighbors[PORTAL] is not None:
                    self.node = self.node.neighbors[PORTAL]
            self.target = self.getNewTarget(direction)
            if self.target is not self.node:
                self.direction = direction
            else:
                self.target = self.getNewTarget(self.direction)

            self.setPosition()
          
    def validDirection(self, direction):
        if direction is not STOP:
            if self.name in self.node.access[direction]:
                if self.node.neighbors[direction] is not None:
                    return True
        return False

    def getNewTarget(self, direction):
        if self.validDirection(direction):
            return self.node.neighbors[direction]
        return self.node

    def overshotTarget(self):
        if self.target is not None:
            vec1 = self.target.position - self.node.position
            vec2 = self.position - self.node.position
            node2Target = vec1.magnitudeSquared()
            node2Self = vec2.magnitudeSquared()
            return node2Self >= node2Target
        return False

    def reverseDirection(self):
        self.direction *= -1
        temp = self.node
        self.node = self.target
        self.target = temp
        
    def oppositeDirection(self, direction):
        if direction is not STOP:
            if direction == self.direction * -1:
                return True
        return False

    def validDirections(self):
        directions = []
        for key in [UP, DOWN, LEFT, RIGHT]:
            if self.validDirection(key):
                if key != self.direction * -1:
                    directions.append(key)
        if len(directions) == 0:
            directions.append(self.direction * -1)
        return directions

    def randomDirection(self, directions):
        return directions[randint(0, len(directions)-1)]

    def goalDirection(self, directions):
        distances = []
        for direction in directions:
            vec = self.node.position  + self.directions[direction]*TILEWIDTH - self.goal
            distances.append(vec.magnitudeSquared())
        index = distances.index(min(distances))
        return directions[index]

    def setStartNode(self, node):
        self.node = node
        self.startNode = node
        self.target = node
        self.setPosition()

    def setBetweenNodes(self, direction):
        if self.node.neighbors[direction] is not None:
            self.target = self.node.neighbors[direction]
            self.position = (self.node.position + self.target.position) / 2.0

    def reset(self):
        self.setStartNode(self.startNode)
        self.direction = STOP
        self.speed = 100
        self.visible = True

    def setSpeed(self, speed):
        self.speed = speed * TILEWIDTH / 16

    def render(self, screen):
        if self.visible:
            if self.image is not None:
                adjust = Vector2(TILEWIDTH, TILEHEIGHT) / 2
                p = self.position - adjust
                screen.blit(self.image, p.asTuple())
            else:
                p = self.position.asInt()
                pygame.draw.circle(screen, self.color, p, self.radius)


from pygame.locals import *
from ..model.constants import *

class LevelManager:
    def __init__(self, game_controller):
        self.gc = game_controller
        self.resetGameState()

    def resetGameState(self):
        self.gc.lives = 5
        self.gc.level = 0
        self.gc.score = 0
        self.gc.lifesprites.resetLives(self.gc.lives)
        self.gc.textgroup.updateScore(self.gc.score)
        self.gc.textgroup.updateLevel(self.gc.level)

    def resetEntities(self):
        self.gc.mouse.reset()
        self.gc.cats.reset()

    def nextLevel(self):
        self.gc.showEntities()
        self.gc.level += 1
        self.gc.pause.paused = True
        self.gc.startGame()
        self.gc.textgroup.updateLevel(self.gc.level)

    def restartGame(self):
        self.gc.pause.paused = True
        self.resetGameState()
        self.gc.startGame()
        self.gc.textgroup.showText(READYTXT)

    def resetLevel(self):
        self.gc.pause.paused = True
        self.resetEntities()
        self.gc.textgroup.showText(READYTXT)



import pygame
from .vector import Vector2
from .constants import *
import numpy as np

class Node(object):
    def __init__(self, x, y):
        self.position = Vector2(x, y)
        self.neighbors = {UP:None, DOWN:None, LEFT:None, RIGHT:None, PORTAL:None}
        self.access = {UP:[MOUSE, CAT1, CAT2, CAT3, CAT4], 
                       DOWN:[MOUSE, CAT1, CAT2, CAT3, CAT4], 
                       LEFT:[MOUSE, CAT1, CAT2, CAT3, CAT4], 
                       RIGHT:[MOUSE, CAT1, CAT2, CAT3, CAT4]}

    def denyAccess(self, direction, entity):
        if entity.name in self.access[direction]:
            self.access[direction].remove(entity.name)

    def allowAccess(self, direction, entity):
        if entity.name not in self.access[direction]:
            self.access[direction].append(entity.name)

    def render(self, screen):
        for n in self.neighbors.keys():
            if self.neighbors[n] is not None:
                line_start = self.position.asTuple()
                line_end = self.neighbors[n].position.asTuple()
                pygame.draw.line(screen, WHITE, line_start, line_end, 4)
                pygame.draw.circle(screen, RED, self.position.asInt(), 12)


class NodeGroup(object):
    def __init__(self, level):
        self.level = level
        self.nodesLUT = {}
        self.nodeSymbols = ['+', 'P', 'n']
        self.pathSymbols = ['.', '-', '|', 'p']
        data = self.readMazeFile(level)
        self.createNodeTable(data)
        self.connectHorizontally(data)
        self.connectVertically(data)
        self.homekey = None

    def readMazeFile(self, textfile):
        return np.loadtxt(textfile, dtype='<U1')

    def createNodeTable(self, data, xoffset=0, yoffset=0):
        for row in list(range(data.shape[0])):
            for col in list(range(data.shape[1])):
                if data[row][col] in self.nodeSymbols:
                    x, y = self.constructKey(col+xoffset, row+yoffset)
                    self.nodesLUT[(x, y)] = Node(x, y)

    def constructKey(self, x, y):
        return x * TILEWIDTH, y * TILEHEIGHT


    def connectHorizontally(self, data, xoffset=0, yoffset=0):
        for row in list(range(data.shape[0])):
            key = None
            for col in list(range(data.shape[1])):
                if data[row][col] in self.nodeSymbols:
                    if key is None:
                        key = self.constructKey(col+xoffset, row+yoffset)
                    else:
                        otherkey = self.constructKey(col+xoffset, row+yoffset)
                        self.nodesLUT[key].neighbors[RIGHT] = self.nodesLUT[otherkey]
                        self.nodesLUT[otherkey].neighbors[LEFT] = self.nodesLUT[key]
                        key = otherkey
                elif data[row][col] not in self.pathSymbols:
                    key = None

    def connectVertically(self, data, xoffset=0, yoffset=0):
        dataT = data.transpose()
        for col in list(range(dataT.shape[0])):
            key = None
            for row in list(range(dataT.shape[1])):
                if dataT[col][row] in self.nodeSymbols:
                    if key is None:
                        key = self.constructKey(col+xoffset, row+yoffset)
                    else:
                        otherkey = self.constructKey(col+xoffset, row+yoffset)
                        self.nodesLUT[key].neighbors[DOWN] = self.nodesLUT[otherkey]
                        self.nodesLUT[otherkey].neighbors[UP] = self.nodesLUT[key]
                        key = otherkey
                elif dataT[col][row] not in self.pathSymbols:
                    key = None


    def getStartTempNode(self):
        nodes = list(self.nodesLUT.values())
        return nodes[0]

    def setPortalPair(self, pair1, pair2):
        key1 = self.constructKey(*pair1)
        key2 = self.constructKey(*pair2)
        if key1 in self.nodesLUT.keys() and key2 in self.nodesLUT.keys():
            self.nodesLUT[key1].neighbors[PORTAL] = self.nodesLUT[key2]
            self.nodesLUT[key2].neighbors[PORTAL] = self.nodesLUT[key1]

    def createHomeNodes(self, xoffset, yoffset):
        homedata = np.array([['X','X','+','X','X'],
                             ['X','X','.','X','X'],
                             ['+','X','.','X','+'],
                             ['+','.','+','.','+'],
                             ['+','X','X','X','+']])

        self.createNodeTable(homedata, xoffset, yoffset)
        self.connectHorizontally(homedata, xoffset, yoffset)
        self.connectVertically(homedata, xoffset, yoffset)
        self.homekey = self.constructKey(xoffset+2, yoffset)
        return self.homekey

    def connectHomeNodes(self, homekey, otherkey, direction):     
        key = self.constructKey(*otherkey)
        self.nodesLUT[homekey].neighbors[direction] = self.nodesLUT[key]
        self.nodesLUT[key].neighbors[direction*-1] = self.nodesLUT[homekey]

    def getNodeFromPixels(self, xpixel, ypixel):
        if (xpixel, ypixel) in self.nodesLUT.keys():
            return self.nodesLUT[(xpixel, ypixel)]
        return None

    def getNodeFromTiles(self, col, row):
        x, y = self.constructKey(col, row)
        if (x, y) in self.nodesLUT.keys():
            return self.nodesLUT[(x, y)]
        return None

    def denyAccess(self, col, row, direction, entity):
        node = self.getNodeFromTiles(col, row)
        if node is not None:
            node.denyAccess(direction, entity)

    def allowAccess(self, col, row, direction, entity):
        node = self.getNodeFromTiles(col, row)
        if node is not None:
            node.allowAccess(direction, entity)

    def denyAccessList(self, col, row, direction, entities):
        for entity in entities:
            self.denyAccess(col, row, direction, entity)

    def allowAccessList(self, col, row, direction, entities):
        for entity in entities:
            self.allowAccess(col, row, direction, entity)

    def denyHomeAccess(self, entity):
        self.nodesLUT[self.homekey].denyAccess(DOWN, entity)

    def allowHomeAccess(self, entity):
        self.nodesLUT[self.homekey].allowAccess(DOWN, entity)

    def denyHomeAccessList(self, entities):
        for entity in entities:
            self.denyHomeAccess(entity)

    def allowHomeAccessList(self, entities):
        for entity in entities:
            self.allowHomeAccess(entity)

    def render(self, screen):
        for node in self.nodesLUT.values():
            node.render(screen)


from pygame.locals import *
from .vector import Vector2
from .constants import *
from .entity import Entity
from .modes import ModeController
from ..view.sprites import CatSprites

class Cat(Entity):
    def __init__(self, node, mouse=None, name = CAT):
        Entity.__init__(self, node)
        self.name = name
        self.points = 200
        self.goal = Vector2()
        self.directionMethod = self.goalDirection
        self.mouse = mouse
        self.mode = ModeController(self)
        self.homeNode = node
        self.sprites = CatSprites(self)

    def reset(self):
        Entity.reset(self)
        self.points = 200
        self.directionMethod = self.goalDirection

    def update(self, dt):
        self.sprites.update(dt)
        self.mode.update(dt)
        if self.mode.current is SCATTER:
            self.scatter()
        elif self.mode.current is CHASE:
            self.chase()
        Entity.update(self, dt)

    def scatter(self):
        self.goal = Vector2()

    def chase(self):
        self.goal = self.mouse.position

    def spawn(self):
        self.goal = self.spawnNode.position

    def setSpawnNode(self, node):
        self.spawnNode = node

    def startSpawn(self):
        self.mode.setSpawnMode()
        if self.mode.current == SPAWN:
            self.setSpeed(150)
            self.directionMethod = self.goalDirection
            self.spawn()

    def startFreight(self):
        self.mode.setFreightMode()
        if self.mode.current == FREIGHT:
            self.setSpeed(50)
            self.directionMethod = self.randomDirection         

    def normalMode(self):
        self.setSpeed(100)
        self.directionMethod = self.goalDirection
        self.homeNode.denyAccess(DOWN, self)
 
class CatGroup(object):
    def __init__(self, node, mouse):
        self.cat1 = Cat(node, mouse, name = CAT1)
        self.cat2 = Cat(node, mouse, name = CAT2)
        self.cat3 = Cat(node, mouse, name = CAT3)
        self.cat4 = Cat(node, mouse, name = CAT4)
        self.cats = [self.cat1, self.cat2, self.cat3, self.cat4]

    def __iter__(self):
        return iter(self.cats)

    def update(self, dt):
        for cat in self:
            cat.update(dt)

    def startFreight(self):
        for cat in self:
            cat.startFreight()
        self.resetPoints()

    def setSpawnNode(self, node):
        for cat in self:
            cat.setSpawnNode(node)

    def updatePoints(self):
        for cat in self:
            cat.points *= 2

    def resetPoints(self):
        for cat in self:
            cat.points = 200

    def hide(self):
        for cat in self:
            cat.visible = False

    def show(self):
        for cat in self:
            cat.visible = True

    def reset(self):
        for cat in self:
            cat.reset()

    def render(self, screen):
        for cat in self:
            cat.render(screen)




import math

class Vector2(object):
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y
        self.thresh = 0.000001

    def __add__(self, other):
        return Vector2(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Vector2(self.x - other.x, self.y - other.y)

    def __neg__(self):
        return Vector2(-self.x, -self.y)

    def __mul__(self, scalar):
        return Vector2(self.x * scalar, self.y * scalar)

    def __div__(self, scalar):
        if scalar != 0:
            return Vector2(self.x / float(scalar), self.y / float(scalar))
        return None

    def __truediv__(self, scalar):
        return self.__div__(scalar)

    def __eq__(self, other):
        if abs(self.x - other.x) < self.thresh:
            if abs(self.y - other.y) < self.thresh:
                return True
        return False

    def magnitudeSquared(self):
        return self.x**2 + self.y**2

    def magnitude(self):
        return math.sqrt(self.magnitudeSquared())

    def copy(self):
        return Vector2(self.x, self.y)

    def asTuple(self):
        return self.x, self.y

    def asInt(self):
        return int(self.x), int(self.y)

    def __str__(self):
        return "<"+str(self.x)+", "+str(self.y)+">"


import logging
from pygame.locals import *
from .controller.game_controller import GameController

logging.basicConfig(level=logging.WARN)
logger = logging.getLogger('CheeseChase')

__all__ = ["GameController"]

# let this be the last line of this file
logger.info("CheeseChase loaded")


import pygame
from ..model.vector import Vector2
from ..model.constants import *

class Text(object):
    def __init__(self, text, color, x, y, size, time=None, id=None, visible=True):
        self.id = id
        self.text = text
        self.color = color
        self.size = size
        self.visible = visible
        self.position = Vector2(x, y)
        self.timer = 0
        self.lifespan = time
        self.label = None
        self.destroy = False
        self.setupFont("CheeseChase/resources/PressStart2P-Regular.ttf")
        self.createLabel()

    def setupFont(self, fontpath):
        self.font = pygame.font.Font(fontpath, self.size)

    def createLabel(self):
        self.label = self.font.render(self.text, 1, self.color)

    def setText(self, newtext):
        self.text = str(newtext)
        self.createLabel()

    def update(self, dt):
        if self.lifespan is not None:
            self.timer += dt
            if self.timer >= self.lifespan:
                self.timer = 0
                self.lifespan = None
                self.destroy = True

    def render(self, screen):
        if self.visible:
            x, y = self.position.asTuple()
            screen.blit(self.label, (x, y))


class TextGroup(object):
    def __init__(self):
        self.nextid = 10
        self.alltext = {}
        self.setupText()
        self.showText(READYTXT)

    def addText(self, text, color, x, y, size, time=None, id=None):
        self.nextid += 1
        self.alltext[self.nextid] = Text(text, color, x, y, size, time=time, id=id)
        return self.nextid

    def removeText(self, id):
        self.alltext.pop(id)
        
    def setupText(self):
        size = TILEHEIGHT
        self.alltext[SCORETXT] = Text("0".zfill(8), RED, 0, TILEHEIGHT, size)
        self.alltext[LEVELTXT] = Text(str(1).zfill(3), RED, 23*TILEWIDTH, TILEHEIGHT, size)
        self.alltext[READYTXT] = Text("READY!", TEAL, 11.25*TILEWIDTH, 20*TILEHEIGHT, size, visible=False)
        self.alltext[PAUSETXT] = Text("PAUSED!", TEAL, 10.625*TILEWIDTH, 20*TILEHEIGHT, size, visible=False)
        self.alltext[GAMEOVERTXT] = Text("GAMEOVER!", TEAL, 10*TILEWIDTH, 20*TILEHEIGHT, size, visible=False)
        self.addText("SCORE", RED, 0, 0, size)
        self.addText("LEVEL", RED, 23*TILEWIDTH, 0, size)

    def update(self, dt):
        for tkey in list(self.alltext.keys()):
            self.alltext[tkey].update(dt)
            if self.alltext[tkey].destroy:
                self.removeText(tkey)

    def showText(self, id):
        self.hideText()
        self.alltext[id].visible = True

    def hideText(self):
        self.alltext[READYTXT].visible = False
        self.alltext[PAUSETXT].visible = False
        self.alltext[GAMEOVERTXT].visible = False

    def updateScore(self, score):
        self.updateText(SCORETXT, str(score).zfill(8))

    def updateLevel(self, level):
        self.updateText(LEVELTXT, str(level + 1).zfill(3))

    def updateText(self, id, value):
        if id in self.alltext.keys():
            self.alltext[id].setText(value)

    def render(self, screen):
        for tkey in list(self.alltext.keys()):
            self.alltext[tkey].render(screen)



import pygame
from pygame.locals import *
from .constants import *
from .entity import Entity
from ..view.sprites import MouseSprites

class Mouse(Entity):
    def __init__(self, node):
        Entity.__init__(self, node )
        self.name = MOUSE    
        self.color = YELLOW
        self.direction = LEFT
        self.setBetweenNodes(LEFT)
        self.alive = True
        self.sprites = MouseSprites(self)

    def reset(self):
        Entity.reset(self)
        self.direction = LEFT
        self.setBetweenNodes(LEFT)
        self.alive = True
        self.image = self.sprites.getStartImage()
        self.sprites.reset()

    def die(self):
        self.alive = False
        self.direction = STOP

    def update(self, dt):	
        self.sprites.update(dt)
        self.position += self.directions[self.direction]*self.speed*dt
        direction = self.getValidKey()
        if self.overshotTarget():
            self.node = self.target
            if self.node.neighbors[PORTAL] is not None:
                self.node = self.node.neighbors[PORTAL]
            self.target = self.getNewTarget(direction)
            if self.target is not self.node:
                self.direction = direction
            else:
                self.target = self.getNewTarget(self.direction)

            if self.target is self.node:
                self.direction = STOP
            self.setPosition()
        else: 
            if self.oppositeDirection(direction):
                self.reverseDirection()

    def getValidKey(self):
        key_pressed = pygame.key.get_pressed()
        if key_pressed[K_UP]:
            return UP
        if key_pressed[K_DOWN]:
            return DOWN
        if key_pressed[K_LEFT]:
            return LEFT
        if key_pressed[K_RIGHT]:
            return RIGHT
        return STOP  

    def eatCheeses(self, cheeseList):
        for cheese in cheeseList:
            if self.collideCheck(cheese):
                return cheese
        return None    
    
    def collideCat(self, cat):
        return self.collideCheck(cat)

    def collideCheck(self, other):
        d = self.position - other.position
        dSquared = d.magnitudeSquared()
        rSquared = (self.collideRadius + other.collideRadius)**2
        if dSquared <= rSquared:
            return True
        return False



from .vector import Vector2
from .constants import *
import numpy as np

class Cheese(object):
    def __init__(self, row, column, spritesheet=None):
        self.name = CHEESE
        self.position = Vector2(column*TILEWIDTH, row*TILEHEIGHT)
        self.collideRadius = 2 * TILEWIDTH / 16
        self.points = 10
        self.visible = True
        self.spritesheet = spritesheet
        self.image = spritesheet.getImage(8, 0)
        
    def render(self, screen):
        if self.visible:
            screen.blit(self.image, self.position.asTuple())

class PowerCheese(Cheese):
    def __init__(self, row, column, spritesheet):
        Cheese.__init__(self, row, column,spritesheet)
        self.name = POWERCHEESE
        self.points = 50
        self.flashTime = 0.4
        self.timer= 0
        self.image = spritesheet.getImage(10,0)
        
    def update(self, dt):
        self.timer += dt
        if self.timer >= self.flashTime:
            self.visible = not self.visible
            self.timer = 0


class CheeseGroup(object):
    def __init__(self, cheesefile, spritesheet=None):
        self.cheeseList = []
        self.powercheeses = []
        self.spritesheet = spritesheet
        self.createCheeseList(cheesefile)
        self.numEaten = 0

    def update(self, dt):
        for powercheese in self.powercheeses:
            powercheese.update(dt)
                
    def createCheeseList(self, cheesefile):
        data = self.readCheesefile(cheesefile)        
        for row in range(data.shape[0]):
            for col in range(data.shape[1]):
                if data[row][col] in ['.', '+']:
                    self.cheeseList.append(Cheese(row, col, self.spritesheet))
                elif data[row][col] in ['P', 'p']:
                    pp = PowerCheese(row, col, self.spritesheet)
                    self.cheeseList.append(pp)
                    self.powercheeses.append(pp)
                    
    def readCheesefile(self, textfile):
        return np.loadtxt(textfile, dtype='<U1')
    
    def isEmpty(self):
        if len(self.cheeseList) == 0:
            return True
        return False
    
    def render(self, screen):
        for cheese in self.cheeseList:
            cheese.render(screen)


import pygame
from ..model.constants import *
import numpy as np
from .animation import Animator

BASETILEWIDTH = 16
BASETILEHEIGHT = 16
DEATH = 5

class Spritesheet(object):
    def __init__(self):
        self.sheet = pygame.image.load("CheeseChase/resources/spritesheet2.png").convert()
        transcolor = self.sheet.get_at((0,0))
        self.sheet.set_colorkey(transcolor)
        
    def getImage(self, x, y, width, height):
        x *= TILEWIDTH
        y *= TILEHEIGHT
        self.sheet.set_clip(pygame.Rect(x, y, width, height))
        return self.sheet.subsurface(self.sheet.get_clip())


class MouseSprites(Spritesheet):
    def __init__(self, entity):
        Spritesheet.__init__(self)
        self.entity = entity
        self.entity.image = self.getStartImage()         
        self.animations = {}
        self.defineAnimations()
        self.stopimage = (0, 0)

    def defineAnimations(self):
        self.animations[DEATH] = Animator(((0, 6), (2, 6), (4, 6), (6, 6), (8, 6), (10, 6), (12, 6), (14, 6), (16, 6), (18, 6)), speed=6, loop=False)

    def update(self, dt):
        if self.entity.alive == True:
            if self.entity.direction == LEFT:
                self.entity.image = self.getImage(0,0)
                self.stopimage = (0, 0)
            elif self.entity.direction == RIGHT:
                self.entity.image = self.getImage(2,0)
                self.stopimage = (2, 0)
            elif self.entity.direction == DOWN:
                self.entity.image = self.getImage(4,0)
                self.stopimage = (4,0)
            elif self.entity.direction == UP:
                self.entity.image = self.getImage(6,0)
                self.stopimage = (6,0)
            elif self.entity.direction == STOP:
                self.entity.image = self.getImage(*self.stopimage)
        else:
            self.entity.image = self.getImage(*self.animations[DEATH].update(dt))

    def reset(self):
        for key in list(self.animations.keys()):
            self.animations[key].reset()

    def getStartImage(self):
        return self.getImage(0, 0)

    def getImage(self, x, y):
        return Spritesheet.getImage(self, x, y, 2*TILEWIDTH, 2*TILEHEIGHT)


class CatSprites(Spritesheet):
    def __init__(self, entity):
        Spritesheet.__init__(self)
        self.entity = entity
        self.entity.image = self.getStartImage()
        self.stopimage = (0, 2)

    def update(self, dt):
        if self.entity.mode.current in [SCATTER, CHASE]:
            if self.entity.direction == LEFT:
                self.entity.image = self.getImage(0, 2)
            elif self.entity.direction == RIGHT:
                self.entity.image = self.getImage(2, 2)
            elif self.entity.direction == DOWN:
                self.entity.image = self.getImage(4, 2)
            elif self.entity.direction == UP:
                self.entity.image = self.getImage(6, 2)
        elif self.entity.mode.current == FREIGHT:
            self.entity.image = self.getImage(8, 2)
        elif self.entity.mode.current == SPAWN:
            if self.entity.direction == LEFT:
                self.entity.image = self.getImage(0, 4)
            elif self.entity.direction == RIGHT:
                self.entity.image = self.getImage(2, 4)
            elif self.entity.direction == DOWN:
                self.entity.image = self.getImage(4, 4)
            elif self.entity.direction == UP:
                self.entity.image = self.getImage(6, 4)
               
    def getStartImage(self):
        return self.getImage(0, 2)

    def getImage(self, x, y):
        return Spritesheet.getImage(self, x, y, 2*TILEWIDTH, 2*TILEHEIGHT)

class LifeSprites(Spritesheet):
    def __init__(self, numlives):
        Spritesheet.__init__(self)
        self.resetLives(numlives)

    def removeImage(self):
        if len(self.images) > 0:
            self.images.pop(0)

    def resetLives(self, numlives):
        self.images = []
        for i in range(numlives):
            self.images.append(self.getImage(0,0))

    def getImage(self, x, y):
        return Spritesheet.getImage(self, x, y, 2*TILEWIDTH, 2*TILEHEIGHT)


class MazeSprites(Spritesheet):
    def __init__(self, mazefile, rotfile):
        Spritesheet.__init__(self)
        self.data = self.readMazeFile(mazefile)
        self.rotdata = self.readMazeFile(rotfile)

    def getImage(self, x, y):
        return Spritesheet.getImage(self, x, y, TILEWIDTH, TILEHEIGHT)

    def readMazeFile(self, mazefile):
        return np.loadtxt(mazefile, dtype='<U1')

    def constructBackground(self, background, y):
        for row in list(range(self.data.shape[0])):
            for col in list(range(self.data.shape[1])):
                if self.data[row][col].isdigit():
                    x = int(self.data[row][col]) + 12
                    sprite = self.getImage(x, y)
                    rotval = int(self.rotdata[row][col])
                    sprite = self.rotate(sprite, rotval)
                    background.blit(sprite, (col*TILEWIDTH, row*TILEHEIGHT))
                #elif self.data[row][col] == '=':
                #    sprite = self.getImage(10, 8)
                #   background.blit(sprite, (col*TILEWIDTH, row*TILEHEIGHT))

        return background

    def rotate(self, sprite, value):
        return pygame.transform.rotate(sprite, value*90)



class Pause(object):
    def __init__(self, paused=False):
        self.paused = paused
        self.timer = 0
        self.pauseTime = None
        self.func = None
        
    def update(self, dt):
        if self.pauseTime is not None:
            self.timer += dt
            if self.timer >= self.pauseTime:
                self.timer = 0
                self.paused = False
                self.pauseTime = None
                return self.func
        return None

    def setPause(self, playerPaused=False, pauseTime=None, func=None):
        self.timer = 0
        self.func = func
        self.pauseTime = pauseTime
        self.flip()

    def flip(self):
        self.paused = not self.paused



#tile and screen settings
TILEWIDTH = 16
TILEHEIGHT = 16
NROWS = 36
NCOLS = 28
SCREENWIDTH = NCOLS*TILEWIDTH
SCREENHEIGHT = NROWS*TILEHEIGHT
SCREENSIZE = (SCREENWIDTH, SCREENHEIGHT)

#usable colors
BLACK = (0, 0, 0)
YELLOW = (255, 255, 0)
WHITE = (255, 255, 255)
RED = (255, 0, 0)
PINK = (255,100,150)
TEAL = (100,255,255)
ORANGE = (230,190,40)
GREEN = (0, 255, 0)

#directions
STOP = 0
UP = 1
DOWN = -1
LEFT = 2
RIGHT = -2
PORTAL = 3

#game objects
MOUSE = 0
CHEESE = 1
POWERCHEESE = 2
CAT = 3
CAT1 = 4
CAT2 = 5
CAT3 = 6
CAT4 = 7

#cat's modes
SCATTER = 0
CHASE = 1
FREIGHT = 2
SPAWN = 3

#texts
SCORETXT = 0
LEVELTXT = 1
READYTXT = 2
PAUSETXT = 3
GAMEOVERTXT = 4



import pygame
from pygame.locals import *
from ..model.constants import *
from .levels_manager import LevelManager

class EventsManager:
    def __init__(self, game_controller):
        self.gc = game_controller  
        self.level_manager = LevelManager(game_controller)

    def checkEvents(self):
        for event in pygame.event.get():
            if event.type == QUIT:
                exit()
            elif event.type == KEYDOWN:
                if event.key == K_SPACE:
                    if self.gc.mouse.alive:
                        self.gc.pause.setPause(playerPaused=True)
                        if not self.gc.pause.paused:
                            self.gc.textgroup.hideText()
                            self.gc.showEntities()
                        else:
                            self.gc.textgroup.showText(PAUSETXT)

    def checkCheeseEvents(self):
        cheese = self.gc.mouse.eatCheeses(self.gc.cheeses.cheeseList)
        if cheese:
            self.gc.cheeses.numEaten += 1
            self.gc.updateScore(cheese.points)
            if self.gc.cheeses.numEaten == 30:
                self.gc.cats.cat3.startNode.allowAccess(RIGHT, self.gc.cats.cat3)
            if self.gc.cheeses.numEaten == 70:
                self.gc.cats.cat4.startNode.allowAccess(LEFT, self.gc.cats.cat4)
            self.gc.cheeses.cheeseList.remove(cheese)
            if cheese.name == POWERCHEESE:
                self.gc.cats.startFreight()
            if self.gc.cheeses.isEmpty():
                self.gc.flashBG = True
                self.gc.hideEntities()
                self.gc.pause.setPause(pauseTime=3, func=self.gc.level_manager.nextLevel)

    def checkCatEvents(self):
        for cat in self.gc.cats:
            if self.gc.mouse.collideCat(cat):
                if cat.mode.current is FREIGHT:
                    self.gc.mouse.visible = False
                    cat.visible = False
                    self.gc.updateScore(cat.points)
                    self.gc.textgroup.addText(str(cat.points), RED, cat.position.x, cat.position.y, 8, time=1)
                    self.gc.cats.updatePoints()
                    self.gc.pause.setPause(pauseTime=1, func=self.gc.showEntities)
                    cat.startSpawn()
                    self.gc.nodes.allowHomeAccess(cat)
                elif cat.mode.current is not SPAWN:
                    if self.gc.mouse.alive:
                        self.gc.lives -= 1
                        self.gc.lifesprites.removeImage()
                        self.gc.mouse.die()
                        self.gc.cats.hide()
                        if self.gc.lives <= 0:
                            self.gc.textgroup.showText(GAMEOVERTXT)
                            self.gc.pause.setPause(pauseTime=3, func=self.gc.level_manager.restartGame)
                        else:
                            self.gc.pause.setPause(pauseTime=3, func=self.gc.level_manager.resetLevel)

